import stripe
from datetime import datetime, timedelta
from enum import Enum
from metasrht.audit import audit_log
from metasrht.types import User, PaymentStatus, Invoice
from metasrht.types import Subscription, SubscriptionStatus
from metasrht.types import PaymentInterval, PaymentCurrency
from srht.config import cfg
from srht.database import db
from srht.graphql import exec_gql

stripe.api_key = cfg("meta.sr.ht::billing", "stripe-secret-key")

class ChargeResult(Enum):
    success = "success"
    failed = "failed"
    cancelled = "cancelled"
    delinquent = "delinquent"
    account_current = "account_current"

def charge_user(user, sub):
    if user.payment_status == PaymentStatus.free:
        return ChargeResult.account_current, "Your account is exempt from payment."

    if user.payment_due >= datetime.utcnow():
        return ChargeResult.account_current, "Your account is current."

    if user.payment_status == PaymentStatus.subsidized:
        user.payment_status = PaymentStatus.unpaid
        return ChargeResult.cancelled, "Your subsidized service term has ended. Contact support for renewal."

    assert sub is not None
    assert sub.currency == PaymentCurrency.USD # TODO

    resp = exec_gql("meta.sr.ht", """
    query {
        me {
            subscription {
                subtotal
                product { id }
            }
        }
    }
    """, user=user)
    amount = resp["me"]["subscription"]["subtotal"]

    print(amount, sub.interval)
    # Temporary guardrails to ensure that we don't accidentally overcharge
    # users during the billing system upgrades
    if sub.interval == PaymentInterval.monthly:
        assert amount in [200, 500, 1000]
    elif sub.interval == PaymentInterval.annually:
        assert amount in [2000, 5000, 10000]

    interval = "monthly"
    if sub.interval == PaymentInterval.annually:
        interval = "annual"
    desc = f"{cfg('sr.ht', 'site-name')} {interval} payment"

    if not sub.autorenew:
        # They cancelled their payment and their current term is up
        sub.status = SubscriptionStatus.INACTIVE
        user.payment_status = PaymentStatus.unpaid
        return ChargeResult.cancelled, "Your subscription has been cancelled."

    try:
        charge = stripe.Charge.create(
            amount=amount,
            currency="usd",
            customer=user.payment_processor_id,
            description=desc)
        audit_log("billing",
                details="charged ${:.2f}".format(amount / 100))
    except stripe.error.CardError as e:
        details = e.json_body["error"]["message"]
        if user.payment_status == PaymentStatus.delinquent:
            # Don't email them twice
            return ChargeResult.delinquent, "Your account payment is delinquent"
        user.payment_status = PaymentStatus.delinquent
        return ChargeResult.failed, details
    except:
        return ChargeResult.failed, "Your payment failed. Contact support."

    invoice = Invoice()
    invoice.user_id = user.id
    invoice.invoice_no = user.next_invoice_no
    user.next_invoice_no += 1

    invoice.product_id = sub.product_id
    invoice.interval = sub.interval

    invoice.currency = sub.currency
    invoice.subtotal = amount
    invoice.tax_charged = 0
    invoice.total = amount
    invoice.payment_id = charge.stripe_id

    db.session.add(invoice)

    invoice.issued = datetime.utcnow()
    invoice.service_start = invoice.issued

    if sub.interval == PaymentInterval.monthly:
        invoice.service_end = datetime.utcnow() + timedelta(days=30)
        user.payment_due = invoice.service_end
    else:
        invoice.service_end = datetime.utcnow()
        leapday = (invoice.service_end.month == 2
                and invoice.service_end.day == 29)
        invoice.service_end = datetime(
                year=invoice.service_end.year + 1,
                month=invoice.service_end.month,
                day=(invoice.service_end.day - 1 if leapday else invoice.service_end.day))
        user.payment_due = invoice.service_end

    user.payment_status = PaymentStatus.current
    db.session.commit()
    return ChargeResult.success, amount
