import requests
import stripe
from datetime import datetime, timedelta
from flask import Blueprint, render_template, request, redirect
from flask import url_for, abort, Response
from iso3166 import countries
from metasrht.audit import audit_log
from metasrht.billing import charge_user
from metasrht.graphql import gql_product, gql_invoice, gql_subscription
from metasrht.types import Subscription, SubscriptionStatus
from metasrht.types import PaymentInterval, PaymentCurrency
from metasrht.types import User, UserType, PaymentStatus, Invoice
from metasrht.webhooks import deliver_profile_update
from sqlalchemy import and_
from srht.config import cfg, get_origin
from srht.crypto import encrypt_request_authorization
from srht.database import db
from srht.flask import session
from srht.graphql import exec_gql, gql_time
from srht.oauth import current_user, loginrequired, freshen_user
from srht.validation import Validation

billing = Blueprint('billing', __name__)
onboarding_redirect = cfg("meta.sr.ht::settings", "onboarding-redirect")

MAX_INVOICE = 4

class BillingState(dict):
    def __init__(self, product_id=None):
        self.product_id = product_id

    def store(self):
        session["billing_state"] = self.__dict__

@billing.route("/billing")
@loginrequired
def billing_GET():
    resp = exec_gql("meta.sr.ht", """
    query BillingInfo {
        me {
            subscription {
                status
                interval
                autorenew
                subtotal
            }

            billingAddress {
                fullName
                businessName
                address1
                address2
                city
                region
                postcode
                country
                vat
            }

            invoices {
                results {
                    id
                    invoiceNo
                    issued
                    serviceEnd
                    total
                }
            }
        }
    }
    """)

    sub = gql_subscription(resp["me"]["subscription"])
    if not sub:
        return redirect(url_for(".billing_setup_GET"))
    invoices = [gql_invoice(i) for i in resp["me"]["invoices"]["results"]]
    if len(invoices) > MAX_INVOICE:
        invoices = invoices[:MAX_INVOICE]
    address = resp["me"]["billingAddress"]
    if address is not None and address.get("country", None) is not None:
        address["country"] = countries.get(address["country"])

    customer = stripe.Customer.retrieve(current_user.payment_processor_id)
    methods = stripe.Customer.list_payment_methods(customer.id)

    return render_template("billing.html",
            message=session.pop("message", None),
            sub=sub, customer=customer, methods=methods,
            invoices=invoices, address=address)

billing_address_query = """
query BillingAddress {
    me {
        billingAddress {
            fullName
            businessName
            address1
            address2
            city
            region
            postcode
            country
            vat
        }
    }
}
"""

@billing.route("/billing/address")
@loginrequired
def billing_address_GET():
    resp = exec_gql("meta.sr.ht", billing_address_query)
    address = resp["me"]["billingAddress"]
    return render_template("billing-edit-address.html",
        address=address, countries=countries)

@billing.route("/billing/address", methods=["POST"])
@loginrequired
def billing_address_POST():
    valid = Validation(request)
    address = {}

    address["country"] = valid.require("country")
    valid.expect(countries.get(address["country"], None) is not None,
            "Invalid country", field="country")
    address["fullName"] = valid.optional("fullName") or None
    address["businessName"] = valid.optional("businessName") or None
    address["address1"] = valid.optional("address1") or None
    address["address2"] = valid.optional("address2") or None
    address["city"] = valid.optional("city") or None
    address["region"] = valid.optional("region") or None
    address["postcode"] = valid.optional("postcode") or None
    address["vat"] = valid.optional("vat") or None

    if not valid.ok:
        resp = exec_gql("meta.sr.ht", billing_address_query)
        address = resp["me"]["billingAddress"]
        return render_template("billing-edit-address.html",
            address=address, countries=countries, valid=valid)

    resp = exec_gql("meta.sr.ht", """
    mutation UpdateAddress($address: BillingAddressInput!) {
        updateBillingAddress(input: $address) {
            country
        }
    }
    """, address=address, valid=valid)

    if not valid.ok:
        resp = exec_gql("meta.sr.ht", billing_address_query)
        address = resp["me"]["billingAddress"]
        return render_template("billing-edit-address.html",
            address=address, countries=countries, valid=valid)

    return redirect(url_for("billing.billing_GET"))

@billing.route("/billing/setup")
@loginrequired
def billing_setup_GET():
    return_to = request.args.get("return_to")
    if return_to:
        session["return_to"] = return_to

    resp = exec_gql("meta.sr.ht", """
    query BillingSetup {
        products {
            id
            name
            prices {
                currency
                amount
            }
        }

        me {
            invoices {
                results {
                    id
                    invoiceNo
                    issued
                    serviceEnd
                    total
                }
            }
        }
    }
    """)

    products = [gql_product(p) for p in resp["products"]]
    invoices = [gql_invoice(i) for i in resp["me"]["invoices"]["results"]]
    if len(invoices) > MAX_INVOICE:
        invoices = invoices[:MAX_INVOICE]

    return render_template("billing-setup.html",
        message=session.pop("message", None),
        invoices=invoices, products=products)

@billing.route("/billing/setup", methods=["POST"])
@loginrequired
def billing_setup_POST():
    valid = Validation(request)

    product_id = valid.require("product_id")
    with valid.catch(ValueError, "Invalid product ID", field="product_id"):
        product_id = int(product_id) if product_id else None
    if not valid.ok:
        return valid.response

    resp = exec_gql("meta.sr.ht", """
    query BillingSetup($productID: Int!) {
        product(id: $productID) {
            id
            name
            retired
        }

        me {
            subscription { id }
        }
    }
    """, productID=product_id)

    sub = gql_subscription(resp["me"]["subscription"])
    product = gql_product(resp["product"])
    valid.expect(product is not None, "Invalid product ID", field="product_id")
    valid.expect(not product or not product["retired"],
                 "This product has been retired",
                 field="product_id")
    if not valid.ok:
        return valid.response

    BillingState(product_id).store()

    if sub is not None:
        return redirect(url_for("billing.billing_chperiod_GET"))

    return redirect(url_for("billing.new_payment_GET"))

@billing.route("/billing/new-payment")
@loginrequired
def new_payment_GET():
    state = session.get("billing_state")
    if state is not None:
        state = BillingState(**state)

    resp = exec_gql("meta.sr.ht", """
    query NewPaymentInfo($productID: Int!) {
        product(id: $productID) { ... productDetails }

        me {
            subscription {
                autorenew
                interval

                product { ... productDetails }
            }
        }
    }

    fragment productDetails on Product {
        id

        prices {
            currency
            amount
        }
    }
    """, productID=state.product_id if state is not None else -1)
    sub = gql_subscription(resp["me"]["subscription"])

    if sub is not None:
        product = sub["product"]
        BillingState(product["id"]).store()
    elif state is not None:
        assert resp["product"] is not None
        product = gql_product(resp["product"])
    else:
        return redirect(url_for("billing.billing_setup_GET"))

    price = product["prices"]["USD"]

    return render_template("new-payment.html",
        product=product, price=price, sub=sub)

@billing.route("/billing/new-payment", methods=["POST"])
@loginrequired
def new_payment_POST():
    valid = Validation(request)
    interval = valid.require("interval", cls=PaymentInterval)
    token = valid.require("stripe-token")
    if not valid.ok:
        return "Invalid form submission", 400

    state = session.get("billing_state")
    state = BillingState(**state)

    resp = exec_gql("meta.sr.ht", """
    query ProductDetails($productID: Int!) {
        product(id: $productID) {
            id

            prices {
                currency
                amount
            }
        }

        # For new-payment.html in case of validation errors:
        me {
            subscription {
                id
                autorenew
                interval
            }
        }
    }
    """, productID=state.product_id)

    sub = gql_subscription(resp["me"]["subscription"])
    product = gql_product(resp["product"])
    assert product is not None
    price = product["prices"]["USD"]
    user = current_user

    # Create/update Stripe customer, add payment method
    if user.payment_processor_id:
        new_customer = False
        try:
            customer = stripe.Customer.retrieve(user.payment_processor_id)
            source = stripe.Customer.create_source(customer.id, source=token)
            customer.default_source = source.stripe_id
            customer.save()
        except stripe.error.CardError as e:
            details = e.json_body["error"]["message"]
            return render_template("new-payment.html",
                product=product, price=price, sub=sub,
                error=details)
    else:
        new_customer = True
        try:
            customer = stripe.Customer.create(
                    description="~" + current_user.username,
                    email=current_user.email,
                    card=token)
            user.payment_processor_id = customer.id
            user.payment_due = datetime.utcnow() + timedelta(minutes=-5)
        except stripe.error.CardError as e:
            details = e.json_body["error"]["message"]
            return render_template("new-payment.html",
                product=product, price=price, sub=sub,
                error=details)

    # Add/update subscription
    sub = (Subscription.query
        .filter(Subscription.user_id == current_user.id)
        .filter(Subscription.status == SubscriptionStatus.ACTIVE)
        .one_or_none())
    if sub:
        sub.product_id = product["id"]
        sub.interval = PaymentInterval(interval)
        sub.autorenew = True

        # Immediately attempt to recharge delinquent users
        if user.payment_status == PaymentStatus.delinquent:
            user.payment_due = datetime.utcnow() + timedelta(minutes=-5)
    else:
        sub = Subscription()
        sub.user_id = user.id
        sub.created = datetime.utcnow()
        sub.updated = datetime.utcnow()
        sub.status = SubscriptionStatus.ACTIVE
        sub.currency = PaymentCurrency.USD
        sub.product_id = product["id"]
        sub.interval = interval
        sub.autorenew = True
        db.session.add(sub)

    audit_log("billing", "New payment method added")
    db.session.commit()

    success, details = charge_user(user, sub)
    db.session.commit()
    if not success:
        return render_template("new-payment.html",
            product=product, price=price, sub=gql_sub,
            error=details)

    freshen_user()

    del session["billing_state"]

    return_to = session.pop("return_to", None)
    if return_to:
        return redirect(return_to)
    if new_customer:
        return redirect(url_for("billing.billing_complete"))
    return redirect(url_for("billing.billing_GET"))

@billing.route("/billing/change-period")
@loginrequired
def billing_chperiod_GET():
    state = session.get("billing_state", None)
    if not state:
        return redirect(url_for("billing.billing_setup_GET"))
    state = BillingState(**state)

    resp = exec_gql("meta.sr.ht", """
    query ChangePeriodDetail($productID: Int!) {
        product(id: $productID) {
            id
            prices {
                currency
                amount
            }
        }

        me {
            subscription {
                autorenew
            }
        }
    }
    """, productID=state.product_id)

    sub = gql_subscription(resp["me"]["subscription"])
    product = gql_product(resp["product"])
    assert product is not None and sub is not None
    price = product["prices"]["USD"]

    return render_template("billing-change-period.html",
        product=product, price=price, sub=sub)

@billing.route("/billing/change-period", methods=["POST"])
def billing_chperiod_POST():
    valid = Validation(request)
    interval = valid.require("interval", cls=PaymentInterval)
    if not valid.ok:
        return "Invalid form submission", 400

    state = session.get("billing_state", None)
    if not state:
        return redirect(url_for("billing.billing_setup_GET"))
    state = BillingState(**state)

    sub = (Subscription.query
        .filter(Subscription.user_id == current_user.id)
        .filter(Subscription.status == SubscriptionStatus.ACTIVE)
        .one())
    sub.interval = interval
    sub.product_id = state.product_id
    sub.autorenew = True
    db.session.commit()

    success, details = charge_user(current_user, sub)
    freshen_user()

    audit_log("billing", "Payment details changed")

    del session["billing_state"]
    return_to = session.pop("return_to", None)
    if return_to:
        return redirect(return_to)
    return redirect(url_for("billing.billing_GET"))

@billing.route("/billing/remove-source/<method_id>", methods=["POST"])
@loginrequired
def payment_source_remove(method_id):
    try:
        stripe.Customer.delete_source(
            current_user.payment_processor_id,
            method_id)
    except stripe.error.StripeError:
        abort(404)
    return redirect(url_for("billing.billing_GET"))

@billing.route("/billing/set-default-source/<method_id>", methods=["POST"])
@loginrequired
def payment_source_make_default(method_id):
    try:
        stripe.Customer.modify(
            current_user.payment_processor_id,
            default_source=method_id)
    except stripe.error.StripeError as ex:
        abort(404)
    return redirect(url_for("billing.billing_GET"))

@billing.route("/billing/complete")
@loginrequired
def billing_complete():
    return render_template("billing-complete.html",
            onboarding_redirect=onboarding_redirect)

@billing.route("/billing/cancel", methods=["POST"])
@loginrequired
def cancel_POST():
    sub = (Subscription.query
        .filter(Subscription.user_id == current_user.id)
        .filter(Subscription.status == SubscriptionStatus.ACTIVE)
        .one_or_none())
    if not sub:
        return redirect(url_for(".billing_GET"))

    sub.autorenew = False

    # If the account is past due, apply the cancellation immediately
    if current_user.payment_status == PaymentStatus.delinquent:
        success, details = charge_user(current_user, sub)
        session["message"] = details
        assert success

    db.session.commit()

    audit_log("billing", "Payment renewal cancelled")
    return redirect(url_for("billing.billing_GET"))

@billing.route("/billing/invoices")
@loginrequired
def invoices_GET():
    cursor = request.args.get("next")

    resp = exec_gql("meta.sr.ht", """
    query Invoices($cursor: Cursor) {
        me {
            invoices(cursor: $cursor) {
                cursor
                results {
                    id
                    invoiceNo
                    issued
                    total
                    serviceEnd
                }
            }
        }
    }
    """, cursor=cursor)

    invoices = [gql_invoice(i) for i in resp["me"]["invoices"]["results"]]
    cursor = resp["me"]["invoices"]["cursor"]

    return render_template("billing-invoices.html",
        invoices=invoices, cursor=cursor)

@billing.route("/billing/invoice/<int:invoice_id>", methods=["POST"])
@loginrequired
def invoice_POST(invoice_id):
    origin = cfg("meta.sr.ht", "api-origin", default=get_origin("meta.sr.ht"))
    headers = {
        "X-Forwarded-For": ", ".join(request.access_route),
        **encrypt_request_authorization(user=current_user),
    }
    r = requests.post(f"{origin}/query/invoice/{invoice_id}",
            headers=headers, data=request.form)
    filename = f"invoice_{invoice_id}.pdf"
    headers = [('Content-Disposition', f'attachment; filename="{filename}"')]
    return Response(r.content, mimetype="application/pdf", headers=headers)
